#!/usr/bin/perl

# Copyright 2002 by Carl Declerck carl@weirdhenge.org

# $Id: bnews-fetch.pl.in,v 1.22 2002/09/08 18:12:24 carl Exp $

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


use IO::File;
use Net::NNTP;
use Getopt::Long;
use strict;

$| = 1;

my $dotdir = (getpwuid($>))[7] . '/.bnews-utils';
my $rcfile = 'fetchrc';
my $opts = &configure($rcfile);

my $decoders = {
	'bn'	=>	'bnews-decoder',
	'yenc'	=>	'ydec -'
};

GetOptions ($opts, 'help!', 'host=s', 'group=s', 'list-queue:s', 'run-queue:s', 'enqueue:s', 'at=s', 'verbosity=s', 'decode:i', 'quiet:i', 'ignore-bad:i', 'from=s', 'to=s', 'range=s', 'fetch!', 'match=s@', 'order-by=s', 'chdir=s', 'pipe-to=s');

defined $opts->{'fetch'} and $opts->{'fetch'} = 1;
defined $opts->{'decode'} and $opts->{'decode'} = 1;
defined $opts->{'quiet'} and $opts->{'quiet'} = 1;
defined $opts->{'ignore-bad'} and $opts->{'ignore-bad'} = 1;
defined $opts->{'enqueue'} and $opts->{'enqueue'} ||= 'default';
defined $opts->{'run-queue'} and $opts->{'run-queue'} ||= 'default';
defined $opts->{'list-queue'} and $opts->{'list-queue'} ||= 'default';

$opts->{'help'} and &help;
$opts->{'at'} and &at($opts);
$opts->{'enqueue'} and &enqueue($opts);
$opts->{'list-queue'} and &listqueue($opts);
$opts->{'run-queue'} and &runqueue($opts);

$opts->{'group'} or die "You must specify a group!\n";

$opts->{'host'} ||= 'newsbin.telenet.be';
$opts->{'order-by'} ||= 'title';
$opts->{'chdir'} ||= '.';
$opts->{'decode'} and $opts->{'pipe-to'} = 'AUTO';
$opts->{'pipe-to'} and $opts->{'fetch'} = 1;
$opts->{'range'} and my ($idx_from, $idx_to) = split(/-/, $opts->{'range'});
$opts->{'quiet'} and $opts->{'verbosity'} = 0;

defined $opts->{'verbosity'} or $opts->{'verbosity'} = -t STDOUT ? 3 : 2;

$opts->{'group'} =~ s/^a\.b\./alt.binaries./i;

my $debug = 0;
my $nntp = Net::NNTP->new($opts->{'host'}) or die "Can't connect to news server $opts->{'host'}\n";

my @grpinfo = $nntp->group(lc($opts->{'group'})) or die "Can't select group $opts->{'group'}\n";

my ($num, $first, $last, $name) = @grpinfo;

$debug and print "first=$first last=$last num=$num\n";

my $overview = $nntp->overview_fmt;
my $idx = {};

for my $n (0..$#{$overview}) {
	$idx->{$overview->[$n]} = $n;
}

$debug and print "Overview: " . join(",", @{$overview}) . "\n";

my $xover = $nntp->xover("$first-$last");

my $files = {};

for my $n (sort { $a <=> $b } keys %{$xover}) {
	my $ovv = $xover->{$n};
	my $subject = $ovv->[$idx->{'Subject:'}];
	my $msgid = $ovv->[$idx->{'Message-ID:'}];
	my $bytes = $ovv->[$idx->{'Bytes:'}];
	my $date = $ovv->[$idx->{'Date:'}];
	my $info = undef;

	$debug and print "msgid=$msgid, subject=$subject\n";

	if ($subject =~ /^BN\s+/) {
		my ($title, $fileinfo) = split(/\s+-:-\s+/, $subject, 2);

		if ($fileinfo =~ s/\s+\((\d+)\/(\d+)\)$//) {
			$info = {
				'type'		=>	'bn',
				'subject'	=>	$subject,
				'msgnum'	=>	$n,
				'msgid'		=>	$msgid,
				'bytes'		=>	$bytes,
				'date'		=>	$date,
				'file'		=>	$fileinfo,
				'title'		=>	$title,
				'chunk'		=>	$1,
				'chunks'	=>	$2
			};
		}
	} elsif ($subject =~ /^(.*\s+yEnc)\s+"([^"]+)"\s+\((\d+)\/(\d+)\)/ 
		|| $subject =~ /^(.*)\s+"([^"]+)"\s+yEnc\s+\((\d+)\/(\d+)\)/
		|| $subject =~ /^(.*)\s+"([^"]+)"\s+\((\d+)\/(\d+)\).*yEnc/
		|| $subject =~ /^(.*)\s+"([^"]+)"\s+\d+\s+yEnc\s+bytes/) {

		$info = {
				'type'		=>	'yenc',
				'subject'	=>	$subject,
				'msgnum'	=>	$n,
				'msgid'		=>	$msgid,
				'bytes'		=>	$bytes,
				'date'		=>	$date,
				'file'		=>	$2,
				'title'		=>	$1,
				'chunk'		=>	$3 || 1,
				'chunks'	=>	$4 || 1
		};
	}

	$opts->{'match'} && !grep($subject =~ /$_/i, @{$opts->{'match'}}) and next;

	my $file = $info->{'file'};
	my $chunk = $info->{'chunk'} + 0;

	defined $info && $chunk != 0 && !defined $files->{$file}->{$chunk} and $files->{$file}->{$chunk} = $info;

}

my $bad = 0;
my $good = 0;
my @fetch = ();
my $from_seen = 0;
my $to_seen = 0;

for my $file (sort { $files->{$a}->{1}->{$opts->{'order-by'}} cmp $files->{$b}->{1}->{$opts->{'order-by'}} } keys %{$files}) {
	my @found = sort { $a <=> $b} keys %{$files->{$file}};
	my $ok = $files->{$file}->{1}->{'chunks'} == $#found + 1;

	if ($ok) {
		$good++;

		$idx_from && $good < $idx_from and next;
		$idx_to && $good > $idx_to and next;

		if (!$from_seen && $opts->{'from'}) {
			$files->{$file}->{1}->{'subject'} =~ /$opts->{'from'}/i and $from_seen = 1;

			$from_seen or next;
		}

		if ($opts->{'to'}) {
			$to_seen and next;

			$files->{$file}->{1}->{'subject'} =~ /$opts->{'to'}/i and $to_seen = 1;
		}

		push (@fetch, $files->{$file});

		$opts->{'verbosity'} >= 2 and printf "%03d: %s, file=%s, chunks=%d\n", $good, $files->{$file}->{1}->{'title'}, $file, $files->{$file}->{1}->{'chunks'};
	} else {
		$bad++;

		$opts->{'verbosity'} >= 2 and print "BAD: $files->{$file}->{1}->{'title'}, file=$file\n";
	}
}

my $cr = -t STDOUT ? "\r" : "";

if ($opts->{'fetch'}) {
	$bad && !$opts->{'ignore-bad'} and die "$bad bad files found, aborting\n";

	mkdir $opts->{'chdir'}, 0755;
	chdir $opts->{'chdir'} or die "Can't chdir to $opts->{'chdir'}\n";

	for my $f (@fetch) {
		my $type = $f->{1}->{'type'};
		my $title = $f->{1}->{'title'};
		my $chunks = $f->{1}->{'chunks'};
		my $file = $f->{1}->{'file'};
		my @yfiles = ();
		my $bnfile = "$file.$type";
		my $output = $opts->{'pipe-to'} ne '' ? "|$opts->{'pipe-to'}" : ">$bnfile";

		if (-f $file) {
			$opts->{'verbosity'} >= 2 and print "SKIP: $title, file=$file\n";

			next;
		}

		$opts->{'pipe-to'} eq 'AUTO' and $output = "|$decoders->{$type}";

		my $fh = IO::File->new($output) or die "Can't open $output ($!)\n";

		$opts->{'verbosity'} == 2 and print "FETCH: $title, file=$file, chunks=$chunks\n";

		for my $chunk (sort { $a <=> $b} keys %{$f}) {
			my $msgid = $f->{$chunk}->{'msgid'};
			my $msgnum = $f->{$chunk}->{'msgnum'};

			$opts->{'verbosity'} >= 3 and print "${cr}FETCH: $title, file=$file, chunk=$chunk/$chunks\n";

			$nntp->body ($msgnum, $fh) or die "Can't fetch article $msgid (msgnum=$msgnum)\n";
		}

		$fh->close;
	}
}

$nntp->quit;

exit 0;

sub configure {
	my $rcfile = shift;
	my $opts = {};
	my $cfg = "$dotdir/$rcfile";

	-f $cfg or return $opts;

	my $fh = IO::File->new($cfg) or die "Can't open config $cfg ($!)\n";

	for my $line ($fh->getlines) {
		$line =~ /^\s*#/ and next;
		$line =~ s/\r//g; chomp $line;

		my ($k, $v) = split(/\s*=\s*/, $line, 2);

		$opts->{$k} = $v;
	}

	$fh->close;

	$opts;
}

sub at {
	my $opts = shift;
	my $at = $opts->{'at'};

	delete $opts->{'at'};

	my $cmd = join(" ", $0, &mkopts($opts));

	my $fh = IO::File->new("|at $at") or die "Can't open at pipe ($!)\n";
	$fh->print ("$cmd\n");
	$fh->close;

	exit 0;
}

sub enqueue {
	my $opts = shift;
	my $q = $opts->{'enqueue'};
	my $qfile = "$dotdir/queues/$q";

	delete $opts->{'enqueue'};

	my $cmd = join(" ", &mkopts($opts));

	mkdir "$dotdir/queues", 0755;

	my $fh = IO::File->new(">>$qfile") or die "Can't open file for queue $q ($!)\n";
	$fh->print ("$cmd\n");
	$fh->close;

	print "QUEUED: $0 $cmd\n";

	exit 0;
}

sub runqueue {
	my $opts = shift;
	my $q = $opts->{'run-queue'};
	my $qfile = "$dotdir/queues/$q";
	my @jobs = ();

	-f $qfile or exit 0;
	while ( -s $qfile ) {
		my $fh = IO::File->new($qfile, "r+") or die "Can't open file for queue $q ($!)\n";
		my @jobs = $fh->getlines;
		my $cmd = pop @jobs;
		$fh->truncate(0);
		$fh->print (@jobs);
		$fh->close;

		print "RUN: $0 $cmd";
		$fh = IO::File->new("|sh") or die "Can't open sh pipe ($!)\n";
		$fh->print ("$0 $cmd");
		$fh->close;
	}

	-z $qfile and unlink "$dotdir/queues/$q";

	exit 0;
}

sub listqueue {
	my $opts = shift;
	my $q = $opts->{'list-queue'};
	my $qfile = "$dotdir/queues/$q";

	-f $qfile or exit 0;

	my $fh = IO::File->new($qfile) or die "Can't open file for queue $q ($!)\n";
	my @jobs = $fh->getlines;
	$fh->close;

	print map { "JOB: $0 $_" } @jobs;

	exit 0;
}

sub mkopts {
	my $opts = shift;
	my @opts = ();

	for my $opt (sort keys %{$opts}) {
		ref($opts->{$opt}) eq 'ARRAY' and push (@opts, join(" ", map { "--$opt='$_'" } @{$opts->{$opt}}));
		ref($opts->{$opt}) eq '' and push (@opts, "--$opt='$opts->{$opt}'");
	}

	@opts;
}

sub help {
	print <<EOT;

Usage: bnews-fetch.pl [options]
  
Options:
  
  --host=hostname : news server to use (default=newsbin.telenet.be)
  --group=groupname : group to select, this is mandatory
  --match=regexp : only show/fetch posts with subject matching regexp
  --order-by=spec : set show/fetch order (spec=title,file,chunk,msgnum)
  --range=start-end : only show/fetch files with numbers in given range
  --from=regexp : only show/fetch files starting from the post matching regexp
  --to=regexp : only show/fetch files ending at the post matching regexp
  --fetch : fetch files to local disk
  --pipe-to=program : don't save to disk but pipe to program, implies --fetch
	Specify AUTO as program to use the built-in default decoders
  --decode : shortcut for --pipe-to=AUTO
  --chdir=dir : mkdir and chdir to specified directory before fetching
  --ignore-bad : go on fetching even when bad/incomplete files were found
  --verbosity=n : set verbosity level (default=3 on a tty, 2 elsewhere)
  --quiet : be quiet, shortcut for --verbosity=0 (errors are still printed)
  --enqueue[=queue] : enqueue as a job on specified queue (or default queue)
  --run-queue[=queue] : execute jobs on specified queue (or default queue)
  --list-queue[=queue] : list jobs on specified queue (or default queue)
  --at=timespec : schedule execution using at(1) at given timespec
  --help : show this help

Default options are read from \$HOME/.bnews-utils/$rcfile if available.

Examples:

\$ bnews-fetch.pl --group=a.b.gothic --match=erben --pipe-to='bnews-decoder -v'
\$ bnews-fetch.pl --group=a.b.goat --match=beloved --fetch --chdir=/bn --at=3am
\$ bnews-fetch.pl --group=a.b.goat --match=tragedy --range=5-13 --decode
\$ bnews-fetch.pl --group=a.b.engines --match=others --from=part38 --decode
\$ bnews-fetch.pl --group=a.b.goat --match=doom --decode --enqueue
\$ bnews-fetch.pl --run-queue --at=1am

EOT
	exit 0;
}
